

#include "a200_switch.h"
#include "sangoma_interface.h"

//globals:
//mutex for synchronizing access to 'stdout' between the threads
HANDLE	print_mutex = NULL;
HANDLE	rx_mutex = NULL;
HANDLE	tdm_event_mutex = NULL;

char	wanpipe_name[DEV_NAME_LEN];	

//arrays of FXO/FXS interfaces
sangoma_interface *sang_if_fxo[MAX_NUM_OF_FXO];
sangoma_interface *sang_if_fxs[MAX_NUM_OF_FXS];

callback_functions_t callback_functions;

SWITCH_STATE switch_state = SW_ST_RESET;

typedef struct{
	void			*sang_if_ptr;
	RX_DATA_STRUCT	event_data;
}TDM_API_EVENT_THREAD_PARAM;

DWORD TdmApiEventThreadFunc(LPDWORD lpdwParam);

/////////////////////////////////////////////////////////////////////

void cli_out(unsigned int dbg_flag, void *pszFormat, ...)
{
#define MAX_DEBUG_STRING	500

	WaitForSingleObject(print_mutex, 1000);
	if(dbg_flag)
	{	static unsigned char str[MAX_DEBUG_STRING];
		va_list	vaArg;

		str[0] = '\0';

		switch(dbg_flag)
		{
		case INFO:
			//do nothing, just print the message
			break;
		case ERR:
			_snprintf((char*)str, MAX_DEBUG_STRING, "Error:");
			break;
		}

#if 1//set to zero to get debug output
		switch(dbg_flag)
		{
		case DBG_IFACE:
		//case DBG_MAIN:
			goto done;
		}
#endif

		switch(dbg_flag)
		{
		case DBG_IFACE:
			_snprintf((char*)str, MAX_DEBUG_STRING, "sangoma_interface::");
			break;
		case DBG_MAIN:
			_snprintf((char*)str, MAX_DEBUG_STRING, "main::");
			break;
		}

		va_start (vaArg, pszFormat);
		_vsnprintf(	(char *)&str[strlen((char*)str)], 
					MAX_DEBUG_STRING-strlen((char*)str), 
					(char *)pszFormat, vaArg);
		va_end (vaArg);

		printf((char *)str);
	}
done:
	ReleaseMutex(print_mutex);
}

//this is where the actual voice switching between ports is done
int got_rx_data(void *sang_if_ptr, void *rx_data)
{
	WaitForSingleObject(rx_mutex, 1000);

	switch(switch_state)
	{
	case SW_ST_RESET:
		//do nothing
		break;

	case SW_ST_FXS_PORT0_TO_FXS_PORT1_CONNECTED:
		if(sang_if_ptr == sang_if_fxs[PORT0]){
			//got data from FXS PORT0, send it to FXS PORT1
			sang_if_fxs[PORT1]->transmit((TX_RX_DATA_STRUCT*)rx_data);
		}
		if(sang_if_ptr == sang_if_fxs[PORT1]){
			//got data from FXS PORT1, send it to FXS PORT0
			sang_if_fxs[PORT0]->transmit((TX_RX_DATA_STRUCT*)rx_data);
		}
		break;
	
	case SW_ST_FXS_PORT0_TO_FXO_PORT0:
		if(sang_if_ptr == sang_if_fxs[PORT0]){
			//got data from FXS PORT0, send it to FXO PORT0
			sang_if_fxo[PORT0]->transmit((TX_RX_DATA_STRUCT*)rx_data);
		}
		if(sang_if_ptr == sang_if_fxo[PORT0]){
			//got data from FXO PORT0, send it to FXS PORT0
			sang_if_fxs[PORT0]->transmit((TX_RX_DATA_STRUCT*)rx_data);
		}
		break;
	}

	ReleaseMutex(rx_mutex);
	return 0;
}

void got_TdmApiEvent(void *sang_if_ptr, void *event_data)
{
	TDM_API_EVENT_THREAD_PARAM	*param;
	DWORD						dwThreadId;

	WaitForSingleObject(tdm_event_mutex, 1000);

	param = (TDM_API_EVENT_THREAD_PARAM*)malloc(sizeof(TDM_API_EVENT_THREAD_PARAM));
	if(param == NULL){
		cli_out(ERR, "Failed to allocate memory for 'Event Thread parameter'!!\n");
		ReleaseMutex(tdm_event_mutex);
		return;
	}

	memcpy(&param->event_data, event_data, sizeof(RX_DATA_STRUCT));
	param->sang_if_ptr = sang_if_ptr;

	//
	//Handling of Events must be done OUTSIDE of the REAL-TIME Rx thread.
	//
	//Create a special thread for Event hadling.
	//
	if(CreateThread( 
        NULL,                       /* no security attributes        */ 
        0,                          /* use default stack size        */ 
        (LPTHREAD_START_ROUTINE)TdmApiEventThreadFunc, /* thread function     */ 
        param,						/* argument to thread function   */ 
        0,                          /* use default creation flags    */ 
        &dwThreadId					/* returns the thread identifier */ 
		) == NULL){
		cli_out(ERR, "Failed to create 'TdmApiEvent' thread!!\n");
	}
	ReleaseMutex(tdm_event_mutex);
}

DWORD TdmApiEventThreadFunc(LPDWORD lpdwParam)
{ 
	TDM_API_EVENT_THREAD_PARAM	*param;
	RX_DATA_STRUCT				*pEvent;
	sangoma_interface			*sang_if;
	wp_tdm_api_rx_hdr_t			*tdm_api_rx_hdr;
	SWITCH_STATE				old_switch_state;

	WaitForSingleObject(tdm_event_mutex, 1000);

	param = (TDM_API_EVENT_THREAD_PARAM*)lpdwParam;

	pEvent = &param->event_data;
	sang_if = (sangoma_interface*)param->sang_if_ptr;

	tdm_api_rx_hdr = (wp_tdm_api_rx_hdr_t*)pEvent->data;

	//keep old state to compare with new one.
	old_switch_state = switch_state;

	cli_out(DBG_MAIN, "TdmApiEventThreadFunc:: ifname: %s\n", sang_if->device_name);

	switch(tdm_api_rx_hdr->wp_tdm_api_event_type)
	{
	case WP_TDM_EVENT_DTMF:
		cli_out(DBG_MAIN, "DTMF Event: Channel: %d, Digit: %c (Port: %s, Type:%s)!\n",
			tdm_api_rx_hdr->wp_tdm_api_event_channel,
			tdm_api_rx_hdr->wp_tdm_api_event_dtmf_digit,
			(tdm_api_rx_hdr->wp_tdm_api_event_dtmf_port == WAN_EC_CHANNEL_PORT_ROUT)?"ROUT":"SOUT",
			(tdm_api_rx_hdr->wp_tdm_api_event_dtmf_type == WAN_EC_TONE_PRESENT)?"PRESENT":"STOP");		

		if(switch_state != SW_ST_RESET){
			//we handle DTMF digits only in SW_ST_RESET. do nothing.
			break;
		}

		switch(tdm_api_rx_hdr->wp_tdm_api_event_dtmf_digit)
		{
		case '3':
			if(sang_if == sang_if_fxs[PORT0]){
				//user wants to call from FXS PORT0 to FXS PORT1
				if(sang_if_fxs[PORT1]->last_hook_state_event == WAN_EVENT_RXHOOK_ON){

					switch_state = SW_ST_FXS_PORT0_TO_FXS_PORT1_DIALING;

					//stop dial tone on port0
					sang_if_fxs[PORT0]->stop_dial_tone();
					//play ring tone on port0
					sang_if_fxs[PORT0]->start_ring_tone();
					//ring the phone on port1
					sang_if_fxs[PORT1]->start_ringing_phone();
				}else{
					//phone on FXS PORT1 is OFF hook (may be talking to someone :) ) play busy tone.
					sang_if_fxs[PORT0]->start_busy_tone();
				}
			}

			if(sang_if == sang_if_fxs[PORT1]){
				//user wants to call from FXS PORT1 to FXS PORT0
				if(sang_if_fxs[PORT0]->last_hook_state_event == WAN_EVENT_RXHOOK_ON){

					switch_state = SW_ST_FXS_PORT0_TO_FXS_PORT1_DIALING;

					//stop dial tone on port1
					sang_if_fxs[PORT1]->stop_dial_tone();
					//play ring tone on port1
					sang_if_fxs[PORT1]->start_ring_tone();
					//ring the phone on port0
					sang_if_fxs[PORT0]->start_ringing_phone();
				}else{
					//phone on FXS PORT0 is OFF hook (may be talking to someone :) ) play busy tone.
					sang_if_fxs[PORT1]->start_busy_tone();
				}
			}
			break;

		case '9':
			//user wants to dial a number on PSTN - go OFF Hook on FXO PORT0
			sang_if_fxo[PORT0]->fxo_go_off_hook();
			//dial tone will be provided by PSTN, stop OUR dial tone on FXS port0
			sang_if_fxs[PORT0]->stop_dial_tone();

			switch_state = SW_ST_FXS_PORT0_TO_FXO_PORT0;
			break;
		}//switch()
		break;

	case WP_TDM_EVENT_RXHOOK:
		cli_out(DBG_MAIN, "RXHOOK Event: %s! (0x%X)\n", 
			WAN_EVENT_RXHOOK_DECODE(tdm_api_rx_hdr->wp_tdm_api_event_rxhook_state),
			tdm_api_rx_hdr->wp_tdm_api_event_rxhook_state);

		sang_if->last_hook_state_event = tdm_api_rx_hdr->wp_tdm_api_event_rxhook_state;

		if(switch_state != SW_ST_FXS_PORT0_TO_FXO_PORT0){
			if(	sang_if_fxs[PORT0]->last_hook_state_event == WAN_EVENT_RXHOOK_ON &&
				sang_if_fxs[PORT1]->last_hook_state_event == WAN_EVENT_RXHOOK_ON){
				//user hang up both phones
				switch_state = SW_ST_RESET;
			}
		}

		switch(switch_state)
		{
		case SW_ST_RESET:
			if(sang_if->last_hook_state_event == WAN_EVENT_RXHOOK_OFF){
				sang_if->start_dial_tone();
			}else{
				sang_if->reset_interface_state();
			}
			break;

		case SW_ST_FXS_PORT0_TO_FXS_PORT1_DIALING:
			if(	sang_if_fxs[PORT0]->last_hook_state_event == WAN_EVENT_RXHOOK_OFF &&
				sang_if_fxs[PORT1]->last_hook_state_event == WAN_EVENT_RXHOOK_OFF){
				//user dialed '3' and picked up both phones
				switch_state = SW_ST_FXS_PORT0_TO_FXS_PORT1_CONNECTED;
			}
			break;

		case SW_ST_FXS_PORT0_TO_FXS_PORT1_CONNECTED:
			if(sang_if->last_hook_state_event == WAN_EVENT_RXHOOK_ON){

				switch_state = SW_ST_RESET;
				//1-st reset all
				sang_if->reset_interface_state();

				//user hang up this phone start "Congestion" on the other phone
				if(sang_if == sang_if_fxs[PORT0]){
					sang_if_fxs[PORT1]->start_congestion_tone();
				}else{
					sang_if_fxs[PORT0]->start_congestion_tone();
				}
			}
			break;

		case SW_ST_FXS_PORT0_TO_FXO_PORT0:
			if(sang_if->last_hook_state_event == WAN_EVENT_RXHOOK_ON){

				if(sang_if == sang_if_fxs[PORT0]){
					//user hang up phone on FXS PORT0, put FXO PORT0 On Hook.
					switch_state = SW_ST_RESET;
					sang_if_fxo[PORT0]->reset_interface_state();
				}
			}
			break;
		}//switch()
		break;

	case WP_TDM_EVENT_RING_DETECT:
		cli_out(DBG_MAIN, "RING Event: %s! (0x%X)\n",	
			WAN_EVENT_RING_DECODE(tdm_api_rx_hdr->wp_tdm_api_event_ring_state),
			tdm_api_rx_hdr->wp_tdm_api_event_ring_state);

		switch(switch_state)
		{
		case SW_ST_RESET:
			if(sang_if == sang_if_fxo[PORT0]){
				switch(tdm_api_rx_hdr->wp_tdm_api_event_ring_state)
				{
				case WAN_EVENT_RING_PRESENT:
					//someone ringing us from PSTN. Ring the phone on FXS PORT0.
					switch_state = SW_ST_PSTN_IS_RINGING;
					//ring the phone on port0
					sang_if_fxs[PORT0]->start_ringing_phone();
					break;

				case WAN_EVENT_RING_STOP:
					//someone ringing us from PSTN. Ring the phone on FXS PORT0.
					switch_state = SW_ST_RESET;
					//stop ringing the phone on port0
					sang_if_fxs[PORT0]->stop_ringing_phone();
					break;
				}
			}
			break;
		}
		break;

	case WP_TDM_EVENT_RING_TRIP:
		cli_out(DBG_MAIN, "RING TRIP Event: %s! (0x%X)\n", 
			WAN_EVENT_RING_TRIP_DECODE(tdm_api_rx_hdr->wp_tdm_api_event_ring_state),
			tdm_api_rx_hdr->wp_tdm_api_event_ring_state);

		switch(switch_state)
		{
		case SW_ST_FXS_PORT0_TO_FXS_PORT1_DIALING:

			//user dialed '3' and picked up the DIALED phone
			switch_state = SW_ST_FXS_PORT0_TO_FXS_PORT1_CONNECTED;

			if(sang_if == sang_if_fxs[PORT0]){
				//user picked up the phone on FXS PORT0
				//stop playing ring tone on port1
				sang_if_fxs[PORT1]->stop_ring_tone();
				//stop ringing the phone on port0
				sang_if_fxs[PORT0]->stop_ringing_phone();
			}

			if(sang_if == sang_if_fxs[PORT1]){
				//user picked up the phone on FXS PORT1
				//stop playing ring tone on port0
				sang_if_fxs[PORT0]->stop_ring_tone();
				//stop ringing the phone on port1
				sang_if_fxs[PORT1]->stop_ringing_phone();
			}
			break;

		case SW_ST_PSTN_IS_RINGING:
			//PSTN called us, we picked up.
			//stop ringing the phone on port0
			sang_if_fxs[PORT0]->stop_ringing_phone();

			//tell PSTN we've heard the ring and we picked up
			sang_if_fxo[PORT0]->fxo_go_off_hook();

			//and now we can talk to PSTN
			switch_state = SW_ST_FXS_PORT0_TO_FXO_PORT0;

			break;

		}//switch()
		break;

	default:
		cli_out(ERR, "Unknown TDM API Event: %d\n", tdm_api_rx_hdr->wp_tdm_api_event_type);
		break;
	}

	if(old_switch_state != switch_state){
		cli_out(INFO, "Switch State Changed from '%s' to '%s'.\n",
			DECODE_SWITCH_STATE(old_switch_state), DECODE_SWITCH_STATE(switch_state));

	}

	free(lpdwParam);
	ReleaseMutex(tdm_event_mutex);

	return 0;
}

int init()
{
	sangoma_interface	*sang_if;
	char				init_status = 0;
	char				if_name[DEV_NAME_LEN];
	int					i, fxo_cnt, fxs_cnt;

	cli_out(DBG_MAIN, "init::\n");

	memset(sang_if_fxo, 0x00, sizeof(sang_if_fxo));
	memset(sang_if_fxs, 0x00, sizeof(sang_if_fxs));

	callback_functions.got_rx_data = got_rx_data;
	callback_functions.got_TdmApiEvent = got_TdmApiEvent;

	//seaching intefaces only on WANPIPE1
	fxo_cnt = fxs_cnt = 0;
	for(i = 0; i < MAX_NUM_OF_INTERFACES && init_status == 0; i++){

		//Form the Interface Name from Wanpipe Name and Inteface Index (i.e. WANPIPE1_IF0).
		_snprintf(if_name, DEV_NAME_LEN, "%s_IF%d", wanpipe_name, i); 
		sang_if = new sangoma_interface(if_name);
		if(sang_if->init(&callback_functions)){
			init_status = 1;
			break;
		}

		if(sang_if->wanif_conf_struct.media == WAN_MEDIA_FXOFXS){
			switch(sang_if->wanif_conf_struct.sub_media)
			{
			case MOD_TYPE_FXS:
				if(fxs_cnt < 2){
					sang_if_fxs[fxs_cnt] = sang_if;
					fxs_cnt++;
				}
				break;
			case MOD_TYPE_FXO:
				if(fxo_cnt < 2){
					sang_if_fxo[fxo_cnt] = sang_if;
					fxo_cnt++;
				}
				break;
			default:
				cli_out(ERR, "Invalid Media sub-type\t\t: Unknown!!\n");
				init_status = 3;
				break;
			}
		}else{
			init_status = 2;
			break;
		}

	}//for()

	if(fxo_cnt != MAX_NUM_OF_FXO){
		cli_out(ERR, "Invalid number of FXO Ports: %d. Has to be exactly %d\n",
			fxo_cnt, MAX_NUM_OF_FXO);
		init_status = 4;
	}

	if(fxs_cnt != MAX_NUM_OF_FXS){
		cli_out(ERR, "Invalid number of FXS Ports: %d. Has to be exactly %d\n",
			fxs_cnt, MAX_NUM_OF_FXS);
		init_status = 5;
	}

	cli_out(DBG_MAIN, "init:: init_status: %d\n", init_status);

	return init_status;
}

void cleanup()
{
	int					i;

	cli_out(DBG_MAIN, "cleanup::\n");

	for(i = 0; i < MAX_NUM_OF_FXS; i++){

		if(sang_if_fxo[i]){
			delete sang_if_fxo[i];
		}

		if(sang_if_fxs[i]){
			delete sang_if_fxs[i];
		}
	}//for()
}

int start()
{
	int rc = 0, i;

	for(i = 0; i < MAX_NUM_OF_FXS; i++){

		sang_if_fxo[i]->run();
		sang_if_fxs[i]->run();
	}//for()

	return rc;
}

void stop()
{
	int i;

	for(i = 0; i < MAX_NUM_OF_FXS; i++){

		sang_if_fxo[i]->stop();
		sang_if_fxs[i]->stop();
	}//for()
}

int main(int argc, char* argv[])
{
	char	rc = 0;
	HANDLE	current_process = NULL;
	char	user_selection;

	printf("A200.EXE version 1.0\n");

	////////////////////////////////////////////////////////////////////////////
	//initialize Wanpipe Name with the default name
	_snprintf(wanpipe_name, DEV_NAME_LEN, "%s", "WANPIPE1");

	////////////////////////////////////////////////////////////////////////////
	//check, if user provided Wanpipe Name as a command line argument
	if(argc == 2){
		_snprintf(wanpipe_name, DEV_NAME_LEN, "%s", argv[1]);
	}
	printf("Using Wanpipe Name: %s\n", wanpipe_name);

	////////////////////////////////////////////////////////////////////////////
	//initialize PrintToScreen mutex
	print_mutex = CreateMutex(NULL, FALSE, NULL);
	if(print_mutex == NULL){
		printf("Failed to create print_mutex!!!!\n");
		return 1;
	}

	rx_mutex = CreateMutex(NULL, FALSE, NULL);
	if(rx_mutex == NULL){
		printf("Failed to create rx_mutex!!!!\n");
		return 1;
	}

	tdm_event_mutex = CreateMutex(NULL, FALSE, NULL);
	if(tdm_event_mutex == NULL){
		printf("Failed to create tdm_event_mutex!!!!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	current_process = GetCurrentProcess();
	//this program 'talks' direcly to hardware, it has to be real time class.
	if(SetPriorityClass(current_process, REALTIME_PRIORITY_CLASS) == 0){
		cli_out(ERR, "Failed to set program priority!\n!");
		return 1;
	}

	rc = init();
	if(rc){
		cleanup();
		return rc;
	}

	rc = start();
	if(rc){
		cleanup();
		return rc;
	}

	do{
		cli_out(INFO, "Press 'q' to quit the program.\n");
		cli_out(INFO, "Press 's' to read switch state.\n");
		user_selection = tolower(getch());
		switch(user_selection)
		{
		case 'q':
			break;
		case 's':
			cli_out(INFO, "Current Switch State: %s (%d)\n", 
				DECODE_SWITCH_STATE(switch_state), switch_state);
			break;
		default:
			cli_out(INFO, "Invalid command.\n");
		}

	}while(user_selection != 'q');

	stop();
	cleanup();
	return rc;
}

