
#include "sangoma_interface.h"

DWORD RxThreadFunc(LPDWORD lpdwParam);

sangoma_interface::sangoma_interface(char *if_name)
{
	cli_out(DBG_IFACE, "sangoma_interface()\n");

	////////////////////////////////////////////////////////////////////////////
	//initialize Device Name with the default name
	_snprintf(device_name , DEV_NAME_LEN, "\\\\.\\%s", if_name);
	cli_out(INFO, "Using Device Name: %s\n", device_name);

	tx_hdr = (api_tx_hdr_t *)wanpipe_api_cmd.data;//do NOT change this pointer!!

	//////////////////////////////////////////////////////////////////
	hGeneralCommands = INVALID_HANDLE_VALUE;
	hRxTx = INVALID_HANDLE_VALUE;

	hRxThread = NULL;
	hTxThread = NULL;

	terminate_tx_rx_threads = 0;
	last_hook_state_event = WAN_EVENT_RXHOOK_ON; //default state is ON Hook

	//////////////////////////////////////////////////////////////////
	//receive stuff
	rx_frames_count = 0;
	rx_bytes_count = 0;
	//for counting frames with CRC/Abort errors
	bad_rx_frames_count = 0;

	//////////////////////////////////////////////////////////////////
	//transmit stuff
	hTxWaitHandle = NULL;
	tx_bytes_count = 0;
	tx_frames_count = 0;
	tx_test_byte = 0;
	hTxCompleteWaitHandle = NULL;

	//////////////////////////////////////////////////////////////////
	//IOCTL management structures and variables
	protocol_cb_size = sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+1;
	wan_protocol = 0;
	adapter_type = 0;
}

sangoma_interface::~sangoma_interface()
{
	cli_out(DBG_IFACE, "~sangoma_interface()\n");
	cleanup();
}

int sangoma_interface::init(callback_functions_t *callback_functions_ptr)
{
	cli_out(DBG_IFACE, "init()\n");

	memcpy(&callback_functions, callback_functions_ptr, sizeof(callback_functions_t));

	////////////////////////////////////////////////////////////////////////////
	//Open device for general commands
	hGeneralCommands = open_api_device(device_name);
    if (hGeneralCommands == INVALID_HANDLE_VALUE){
		cli_out(ERR, "Unable to open %s for general commands!\n", device_name);
		return 1;
	}

	//open another handle strictly for reading
	hRxTx = open_api_device(device_name);
    if(hRxTx == INVALID_HANDLE_VALUE){
		cli_out(ERR, "\nUnable to open %s for reading!\n", device_name);
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get current protocol
	if(GetWANConfig() == WAN_FALSE){
		cli_out(ERR, "Failed to get current protocol!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//get Front End Type (T1/E1/Analog...)
	if (get_fe_type(&adapter_type) == WAN_FALSE){
		cli_out(ERR, "Failed to get Front End Type!\n");
		return 1;
	}

	////////////////////////////////////////////////////////////////////////////
	//may need interface configuration, so get it now
	if(get_interface_configuration(&wanif_conf_struct)){
		cli_out(ERR, "Failed to get Interface Configuration!\n");
		return 1;
	}

	return 0;
}

HANDLE sangoma_interface::open_api_device(char *api_device_name)
{
	cli_out(DBG_IFACE, "open_api_device()\n");

	return CreateFile(	api_device_name, 
						GENERIC_READ | GENERIC_WRITE, 
						FILE_SHARE_READ | FILE_SHARE_WRITE,
						(LPSECURITY_ATTRIBUTES)NULL, 
						OPEN_EXISTING,
						FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
						(HANDLE)NULL
						);
}

unsigned char sangoma_interface::get_wan_udphdr_data_byte(unsigned char off) 
{
	int i=0; 

	for (;;){

		if (function_lookup[i].protocol_id == 0){
			cli_out(ERR, "\nProtocol: '%s' support not compiled in!\n\n", 
					DECODE_PROT(wan_protocol)); 
			break;
		};
	
		if (function_lookup[i].protocol_id == wan_protocol){

			return (((unsigned char*)&wan_udp.wan_udphdr_data[0])+
				function_lookup[i].mbox_offset)[off];
		}

		i++;
	}

	return 0;
}

int sangoma_interface::GetWANConfig( void )
{
	int err = WAN_TRUE;

	/* Get Protocol type */ 
	wan_udp.wan_udphdr_command	= WAN_GET_PROTOCOL;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: Command WAN_GET_PROTOCOL failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	wan_protocol = wan_udp.wan_udphdr_data[0];

	cli_out(INFO, "Device %s running protocol: %s\n",
		device_name, DECODE_PROT(wan_protocol));
	 
	return err;
} 

int sangoma_interface::get_fe_type(unsigned char* adapter_type)
{
	int err = WAN_TRUE;

	/* Read Adapter Type */
	wan_udp.wan_udphdr_command = WAN_GET_MEDIA_TYPE;
	wan_udp.wan_udphdr_data[0] = WAN_MEDIA_NONE;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: Command WAN_GET_MEDIA_TYPE failed! return code: 0x%X",
			wan_udp.wan_udphdr_return_code);
		return WAN_FALSE;
	}

	*adapter_type =	get_wan_udphdr_data_byte(0);

	cli_out(INFO, "Front End Type: ");
	switch(*adapter_type)
	{
	case WAN_MEDIA_NONE:
		cli_out(INFO, "Serial");
		break;

	case WAN_MEDIA_T1:
		cli_out(INFO, "T1");
		break;

	case WAN_MEDIA_E1:
		cli_out(INFO, "E1");
		break;

	case WAN_MEDIA_56K:
		cli_out(INFO, "56K");
		break;

	case WAN_MEDIA_FXOFXS:
		cli_out(INFO, "Aanalog");
		break;

	default:
		cli_out(INFO, "Unknown");
		err = WAN_FALSE;
	}
	cli_out(INFO, "\n");

	return err;
}

unsigned char* sangoma_interface::get_wan_udphdr_data_ptr(unsigned char off) 
{
	int i=0; 
	for (;;){
		if (function_lookup[i].protocol_id == 0){
			cli_out(ERR, "\nProtocol: '%s' support not compiled in!\n\n", 
					DECODE_PROT(wan_protocol)); 
			break;
		};
	
		if (function_lookup[i].protocol_id == wan_protocol){

			return &(((unsigned char*)&wan_udp.wan_udphdr_data[0])+
				function_lookup[i].mbox_offset)[off];
		}
		i++;
	}

	return 0;
}

unsigned char sangoma_interface::set_wan_udphdr_data_byte(unsigned char off, unsigned char data) 
{
	int i=0; 
	for (;;){
		if (function_lookup[i].protocol_id == 0){
			cli_out(ERR, "\nProtocol: '%s' support not compiled in!\n\n", 
					DECODE_PROT(wan_protocol)); 
			break;
		};
	
		if (function_lookup[i].protocol_id == wan_protocol){

			(((unsigned char*)&wan_udp.wan_udphdr_data[0])+
				function_lookup[i].mbox_offset)[off] = data;

			return 0;
		}
		i++;
	}

	return -1;
}

void sangoma_interface::read_te1_56k_stat(void)
{
	unsigned char* data = NULL;
	unsigned long 	alarms = 0x00;
	sdla_te_pmon_t*	pmon = NULL;

	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;

	default:
		cli_out(ERR, "Command invalid for Adapter Type %d.\n", adapter_type);
		return;
	}

	/* Read T1/E1/56K alarms and T1/E1 performance monitoring counters */
	wan_udp.wan_udphdr_command = WAN_FE_GET_STAT;
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code != 0){
		cli_out(ERR, "Failed to read T1/E1/56K statistics.\n");
		return;
	}

	data = get_wan_udphdr_data_ptr(0);

#if 0
	if (wan_protocol == WANCONFIG_CHDLC){
		data = &wan_udp.wan_udphdr_chdlc_data[0];
	}else{
		data = &wan_udp.wan_udphdr_data[0];
	}
#endif

	alarms = *(unsigned long*)data;
	if (adapter_type == WAN_MEDIA_T1 || adapter_type == WAN_MEDIA_E1){
		cli_out(INFO, "***** %s: %s Alarms *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
		cli_out(INFO, "ALOS:\t%s\t| LOS:\t%s\n", 
				WAN_TE_ALOS_ALARM(alarms), 
				WAN_TE_LOS_ALARM(alarms));
		cli_out(INFO, "RED:\t%s\t| AIS:\t%s\n", 
				WAN_TE_RED_ALARM(alarms), 
				WAN_TE_AIS_ALARM(alarms));
		if (adapter_type == WAN_MEDIA_T1){ 
			cli_out(INFO, "YEL:\t%s\t| OOF:\t%s\n", 
					WAN_TE_YEL_ALARM(alarms), 
					WAN_TE_OOF_ALARM(alarms));
		}else{
			cli_out(INFO, "OOF:\t%s\n", 
					WAN_TE_OOF_ALARM(alarms));
		}
	}else{
		cli_out(INFO, "***** %s: 56K CSU/DSU Alarms *****\n\n\n", device_name);
	 	cli_out(INFO, "In Service:\t\t%s\tData mode idle:\t\t%s\n",
			 	INS_ALARM_56K(alarms), 
			 	DMI_ALARM_56K(alarms));
	
	 	cli_out(INFO, "Zero supp. code:\t%s\tCtrl mode idle:\t\t%s\n",
			 	ZCS_ALARM_56K(alarms), 
			 	CMI_ALARM_56K(alarms));

	 	cli_out(INFO, "Out of service code:\t%s\tOut of frame code:\t%s\n",
			 	OOS_ALARM_56K(alarms), 
			 	OOF_ALARM_56K(alarms));
		
	 	cli_out(INFO, "Valid DSU NL loopback:\t%s\tUnsigned mux code:\t%s\n",
			 	DLP_ALARM_56K(alarms), 
			 	UMC_ALARM_56K(alarms));

	 	cli_out(INFO, "Rx loss of signal:\t%s\t\n",
			 	RLOS_ALARM_56K(alarms)); 
	}

	if (adapter_type != WAN_MEDIA_T1 && adapter_type != WAN_MEDIA_E1){
		return;
	}
	pmon = (sdla_te_pmon_t*)&data[sizeof(unsigned long)];
/*
//FIXME:
	cli_out(INFO, "\n\n***** %s: %s Performance Monitoring Counters *****\n\n",
			device_name, (adapter_type == WAN_MEDIA_T1) ? "T1" : "E1");
	cli_out(INFO, "Framing Bit Error:\t%ld\tLine Code Violation:\t%ld\n", 
			pmon->frm_bit_error,
			pmon->lcv);
	if (adapter_type == WAN_MEDIA_T1){
		cli_out(INFO, "Out of Frame Errors:\t%ld\tBit Errors:\t\t%ld\n", 
				pmon->oof_errors,
				pmon->bit_errors);
	}else{
		cli_out(INFO, "Far End Block Errors:\t%ld\tCRC Errors:\t%ld\n", 
				pmon->far_end_blk_errors,
				pmon->crc_errors);
	}
*/
	return;
}

void sangoma_interface::set_lb_modes(unsigned char type, unsigned char mode)
{
	switch(adapter_type)
	{
	case WAN_MEDIA_T1:
	case WAN_MEDIA_E1:
	case WAN_MEDIA_56K:
		;//do nothing
		break;

	default:
		cli_out(ERR, "Command invalid for Adapter Type %d.\n", adapter_type);
		return;
	}
	
	wan_udp.wan_udphdr_command	= WAN_FE_SET_LB_MODE;
	wan_udp.wan_udphdr_data_len	= 2;

	set_wan_udphdr_data_byte(0,type);
	set_wan_udphdr_data_byte(1,mode);

#if 0
	if (wan_protocol == WANCONFIG_CHDLC){
		wan_udp.wan_udphdr_chdlc_data[0] = type;
		wan_udp.wan_udphdr_chdlc_data[1] = mode;
	}else{
		wan_udp.wan_udphdr_data[0] = type;
		wan_udp.wan_udphdr_data[1] = mode;
	}
#endif
	DO_COMMAND(wan_udp);
	if (wan_udp.wan_udphdr_return_code != 0){
		cli_out(ERR, "Failed to %s line loopback mode. rc: 0x%X\n",
			((mode == WAN_TE1_ACTIVATE_LB) ? "activate" : "deactivate"), 
			wan_udp.wan_udphdr_return_code);
	}else{
		cli_out(INFO, "Line loopback mode is %s!\n",
			(mode == WAN_TE1_ACTIVATE_LB) ? "activated" : "deactivated");
	}
	return;
}

void sangoma_interface::operational_stats (void)
{
	net_device_stats_t *stats;
	wan_udp.wan_udphdr_command = READ_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command READ_OPERATIONAL_STATS failed!\n");
	}

	stats = (net_device_stats_t *)get_wan_udphdr_data_ptr(0);

	cli_out(INFO, "**** OPERATIONAL_STATS ****\n");

	cli_out(INFO, "\ttx_dropped\t: %lu\n", stats->tx_dropped);
	cli_out(INFO, "\ttx_packets\t: %lu\n", stats->tx_packets);
	cli_out(INFO, "\ttx_bytes\t: %lu\n", stats->tx_bytes);
	cli_out(INFO, "\n");

	//receive stats
	cli_out(INFO, "\trx_packets\t: %lu\n", stats->rx_packets);
	cli_out(INFO, "\trx_bytes\t: %lu\n", stats->rx_bytes);

	cli_out(INFO, "\trx_errors\t: %lu\n", stats->rx_errors);
	cli_out(INFO, "\trx_frame_errors\t: %lu\n", stats->rx_frame_errors);

	cli_out(INFO, "\trx_packets_discarded_rx_q_full\t: %lu\n",
		stats->rx_packets_discarded_rx_q_full);
	cli_out(INFO, "\trx_packets_discarded_excessive_length\t: %lu\n",
		stats->rx_packets_discarded_excessive_length);
	
	cli_out(INFO, "\trx_packets_discarded_too_short\t: %lu\n", 
		stats->rx_packets_discarded_too_short);
	cli_out(INFO, "\n");

	//hardware level stats
	//AFT card
	cli_out(INFO, "\trx_fifo_errors\t: %lu\n", stats->rx_fifo_errors);
	cli_out(INFO, "\ttx_fifo_errors\t: %lu\n", stats->tx_fifo_errors);
	//S508, S514
	cli_out(INFO, "\tnumber_of_times_rx_buffers_out_of_synch_with_card\t: %lu\n",
		stats->number_of_times_rx_buffers_out_of_synch_with_card);

	//BitStream mode only - counter of data blocks transmitted automatically,
	//because no User Data was available. Transmit underrun counter from user
	//point of view.
	cli_out(INFO, "\ttx_idle_data\t: %lu\n", stats->tx_idle_data);
}

void sangoma_interface::flush_operational_stats (void)
{
	wan_udp.wan_udphdr_command = FLUSH_OPERATIONAL_STATS; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command FLUSH_OPERATIONAL_STATS failed!\n");
	}

	cli_out(INFO, "Command FLUSH_OPERATIONAL_STATS was successful.\n");
}

//get RBS being received
char sangoma_interface::get_rbs(rbs_management_t *rbs_management_ptr)
{
	wan_udp.wan_udphdr_command = WAN_GET_RBS_BITS;
	memcpy(&wan_udp.wan_udphdr_data[0], rbs_management_ptr, sizeof(rbs_management_t));
	wan_udp.wan_udphdr_data_len = sizeof(rbs_management_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command WAN_GET_RBS_BITS failed!\n");
		return 1;
	}

	rbs_management_ptr = (rbs_management_t *)&wan_udp.wan_udphdr_data[0];

	cli_out(INFO, "**** WAN_GET_RBS_BITS OK ****\n");

	//cli_out(INFO, "ABCD_bits (HEX): 0x%02X\n", rbs_management_ptr->ABCD_bits);

	cli_out(INFO, "Channel: %d, RX RBS: A:%1d B:%1d C:%1d D:%1d\n",
			rbs_management_ptr->channel, 
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_A) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_B) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_C) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_D) ? 1 : 0);
	return 0;
}

//set RBS to be transmitted
char sangoma_interface::set_rbs(rbs_management_t *rbs_management_ptr)
{
	wan_udp.wan_udphdr_command = WAN_SET_RBS_BITS;
	memcpy(&wan_udp.wan_udphdr_data[0], rbs_management_ptr, sizeof(rbs_management_t));
	wan_udp.wan_udphdr_data_len = sizeof(rbs_management_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command WAN_SET_RBS_BITS failed!\n");
		return 1;
	}

	cli_out(INFO, "**** WAN_SET_RBS_BITS OK ****\n");

	//cli_out(INFO, "ABCD_bits (HEX): 0x%02X\n", rbs_management_ptr->ABCD_bits);

	cli_out(INFO, "Channel: %d, TX RBS: A:%1d B:%1d C:%1d D:%1d\n",
			rbs_management_ptr->channel, 
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_A) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_B) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_C) ? 1 : 0,
			(rbs_management_ptr->ABCD_bits & WAN_RBS_SIG_D) ? 1 : 0);
	return 0;
}

int sangoma_interface::get_interface_configuration(if_cfg_t *wanif_conf_ptr)
{
	wan_udp.wan_udphdr_command = READ_CONFIGURATION;
	wan_udp.wan_udphdr_data_len = sizeof(if_cfg_t);

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command READ_CONFIGURATION failed!\n");
		return 1;
	}
	memcpy(wanif_conf_ptr, get_wan_udphdr_data_ptr(0), sizeof(if_cfg_t));

	cli_out(INFO, "**** READ_CONFIGURATION ****\n");
	cli_out(INFO, "Operational Mode\t: %s\n", wanif_conf_ptr->usedby);
	cli_out(INFO, "Active Channels \t: 0x%08X\n", wanif_conf_ptr->active_ch);
	cli_out(INFO, "Interface Number\t: %u\n", wanif_conf_ptr->interface_number);
	cli_out(INFO, "Media type\t\t: %u\n", wanif_conf_ptr->media);
	if(wanif_conf_ptr->media == WAN_MEDIA_FXOFXS){
		switch(wanif_conf_ptr->sub_media)
		{
		case MOD_TYPE_FXS:
			cli_out(INFO, "Media sub-type\t\t: FXS\n");
			break;
		case MOD_TYPE_FXO:
			cli_out(INFO, "Media sub-type\t\t: FXO\n");
			break;
		default:
			cli_out(INFO, "Media sub-type\t\t: Unknown!!\n");
			break;
		}
	}
	cli_out(INFO, "****************************\n");

	return 0;
}

int sangoma_interface::get_open_handles_counter()
{
	wan_udp.wan_udphdr_command = GET_OPEN_HANDLES_COUNTER; 
	wan_udp.wan_udphdr_data_len = 0;

	DO_COMMAND(wan_udp);
	if(wan_udp.wan_udphdr_return_code){
		cli_out(ERR, "Error: command GET_OPEN_HANDLES_COUNTER failed!\n");
		return -1;
	}

	return *(int*)&wan_udp.wan_udphdr_data[0];
}


//It is very important to close ALL open
//handles, because the API will continue
//receive data until the LAST handel is closed.
void sangoma_interface::cleanup()
{
	cli_out(INFO, "sangoma_interface::cleanup()\n");

	if(hGeneralCommands != INVALID_HANDLE_VALUE){
		int hndls_counter = get_open_handles_counter();
		if(hndls_counter > 0){
			cli_out(INFO, "There are %d API Handles to be closed.\n", hndls_counter);
		}
		cli_out(INFO, "Closing hGeneralCommands.\n");
		CloseHandle(hGeneralCommands);
		hGeneralCommands = INVALID_HANDLE_VALUE;
	}

	if(hRxTx != INVALID_HANDLE_VALUE){
		cli_out(INFO, "Closing hRxTx.\n");
		CloseHandle(hRxTx);
		hRxTx = INVALID_HANDLE_VALUE;
	}

	if(hTxWaitHandle != NULL){
		CloseHandle(hTxWaitHandle);
		hTxWaitHandle = NULL;
	}

	if(hTxCompleteWaitHandle != NULL){
		CloseHandle(hTxCompleteWaitHandle);
		hTxCompleteWaitHandle = NULL;
	}
}

// Blocking write command.
UCHAR sangoma_interface::DoWriteCommand(HANDLE drv, TX_DATA_STRUCT * pTx)
{
	DWORD ln;

	if(DeviceIoControl(
			drv,
			IoctlWriteCommand,
			(LPVOID)pTx,
			(ULONG)sizeof(TX_DATA_STRUCT),
			(LPVOID)pTx,
			sizeof(TX_DATA_STRUCT),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

// Blocking read command.
USHORT sangoma_interface::DoReadCommand(HANDLE drv, RX_DATA_STRUCT * pRx)
{
	DWORD ln;

	if (DeviceIoControl(
			drv,
			IoctlReadCommand,
			(LPVOID)NULL,
			0L,
			(LPVOID)pRx,
			sizeof(RX_DATA_STRUCT),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

// set data in Idle Transmit buffer (BitStream only).
UCHAR sangoma_interface::DoSetIdleTxBufferCommand(HANDLE drv, TX_DATA_STRUCT * pTx)
{
	DWORD ln;

	if(DeviceIoControl(
			drv,
			IoctlSetIdleTxBuffer,
			(LPVOID)pTx,
			(ULONG)sizeof(TX_DATA_STRUCT),
			(LPVOID)pTx,
			sizeof(TX_DATA_STRUCT),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

int sangoma_interface::DoManagementCommand(HANDLE drv, wan_udp_hdr_t* wan_udp)
{
	DWORD ln;

    static unsigned char id = 0;
	int data_len = protocol_cb_size+wan_udp->wan_udphdr_data_len;

	wan_udp->wan_udphdr_request_reply = 0x01;
	wan_udp->wan_udphdr_id = id;
   	wan_udp->wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	if(DeviceIoControl(
			drv,
			IoctlManagementCommand,
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
						) == FALSE){
		//check messages log
		return 1;
	}else{
		return 0;
	}
}

int sangoma_interface::tdmv_api_ioctl(wanpipe_tdm_api_cmd_t *api_cmd, int cmd)
{
	DWORD			ln;
    unsigned char	id = 0;
	int				err = 0;

	wan_udp.wan_udphdr_request_reply = 0x01;
	wan_udp.wan_udphdr_id = id;
   	wan_udp.wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	wan_udp.wan_udphdr_command	= WAN_TDMV_API_IOCTL;
	wan_udp.wan_udphdr_data_len	= sizeof(wanpipe_tdm_api_cmd_t);

	api_cmd->cmd = cmd;
	//copy data from caller's buffer to driver's buffer
	memcpy(	wan_udp.wan_udphdr_data, 
			(void*)api_cmd,
			sizeof(wanpipe_tdm_api_cmd_t));

	if(DeviceIoControl(
			hGeneralCommands,
			IoctlManagementCommand,
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
			) == FALSE){
		err = 1;
	}else{
		err = 0;
	}

	if(err){
		printf("Error: tdmv_api_ioctl(): DeviceIoControl failed!!\n");
		return 1;
	}
	if(wan_udp.wan_udphdr_return_code != WAN_CMD_OK){
		printf("Error: tdmv_api_ioctl(): command failed! Return code: 0x%X.\n",
			wan_udp.wan_udphdr_return_code);
		return 2;
	}

	//copy data from driver's buffer to caller's buffer
	memcpy(	(void*)api_cmd,
			wan_udp.wan_udphdr_data, 
			sizeof(wanpipe_tdm_api_cmd_t));

	//delay AFTER the command!!
	EVENT_DELAY();

	return 0;
}

int sangoma_interface::wanpipe_api_ioctl(wan_cmd_api_t *api_cmd)
{
	DWORD			ln;
    unsigned char	id = 0;
	int				err = 0;

	wan_udp.wan_udphdr_request_reply = 0x01;
	wan_udp.wan_udphdr_id = id;
   	wan_udp.wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;

	wan_udp.wan_udphdr_command	= SIOC_WANPIPE_API;
	wan_udp.wan_udphdr_data_len	= sizeof(wan_cmd_api_t);

	//copy data from caller's buffer to driver's buffer
	memcpy(	wan_udp.wan_udphdr_data, 
			(void*)api_cmd,
			sizeof(wan_cmd_api_t));

	if(DeviceIoControl(
			hGeneralCommands,
			IoctlManagementCommand,
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPVOID)&wan_udp,
			sizeof(wan_udp_hdr_t),
			(LPDWORD)(&ln),
			(LPOVERLAPPED)NULL
			) == FALSE){
		err = 1;
	}else{
		err = 0;
	}

	if(err){
		printf("Error: wanpipe_api_ioctl(): DeviceIoControl failed!!\n");
		return 1;
	}
	if(wan_udp.wan_udphdr_return_code != WAN_CMD_OK){
		printf("Error: wanpipe_api_ioctl(): command failed! Return code: 0x%X.\n",
			wan_udp.wan_udphdr_return_code);
		return 2;
	}

	//copy data from driver's buffer to caller's buffer
	memcpy(	(void*)api_cmd,
			wan_udp.wan_udphdr_data, 
			sizeof(wan_cmd_api_t));

	//delay AFTER the command!!
	EVENT_DELAY();

	return 0;
}


int sangoma_interface::stop()
{
	cli_out(DBG_IFACE, "stop()\n");

	cli_out(INFO, "Stopping.");

	terminate_tx_rx_threads = 1;
	while(terminate_tx_rx_threads == 1){
		cli_out(INFO, ".");
		Sleep(500);
	}
	cli_out(INFO, "\n");

	if(wanif_conf_struct.sub_media == MOD_TYPE_FXS){
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_DISABLE_RXHOOK_EVENTS);
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_DISABLE_RM_DTMF_EVENTS);
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_DISABLE_RING_TRIP_DETECT_EVENTS);
	}

	if(wanif_conf_struct.sub_media == MOD_TYPE_FXO){
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_DISABLE_RING_DETECT_EVENTS);
	}

	return 0;
}

int sangoma_interface::run()
{
	cli_out(DBG_IFACE, "run()\n");

	if(wanif_conf_struct.sub_media == MOD_TYPE_FXS){
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_ENABLE_RXHOOK_EVENTS);
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_ENABLE_RM_DTMF_EVENTS);
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_ENABLE_RING_TRIP_DETECT_EVENTS);
	}
	if(wanif_conf_struct.sub_media == MOD_TYPE_FXO){
		tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_ENABLE_RING_DETECT_EVENTS);
	}

	////////////////////////////////////////////////////////////////////////////
	//Start a thread for receiving data only
    hRxThread = CreateThread( 
        NULL,                       /* no security attributes        */ 
        0,                          /* use default stack size        */ 
        (LPTHREAD_START_ROUTINE)RxThreadFunc, /* thread function     */ 
        this,						/* argument to thread function   */ 
        0,                          /* use default creation flags    */ 
        &dwRxThreadId				/* returns the thread identifier */ 
		);

	if(hRxThread == NULL){
		cli_out(ERR, "Failed to create Rx thread!!\n");
		return 1;
	}

	if(SetThreadPriority(hRxThread, THREAD_PRIORITY_TIME_CRITICAL) == FALSE){
		cli_out(ERR, "Failed to set Rx thread priority!!\n");
		return 1;
	}

	return 0;
}

// Read Thread
DWORD RxThreadFunc(LPDWORD lpdwParam)
{ 
	api_header_t		*pri;
	sangoma_interface	*sang_if = (sangoma_interface*)lpdwParam;

    cli_out(INFO, "\n%s: RxThreadFunc() - start\n", sang_if->device_name); 

	while(sang_if->terminate_tx_rx_threads == 0){	

		if(sang_if->receive(&sang_if->rx_data) ){
			//error
			cli_out(ERR, "receive() failed!! Check messages log.\n");
			break;
		}

		//use our special buffer at rxdata to hold received data
		pri = &sang_if->rx_data.api_header;

		switch(pri->operation_status)
		{
		case SANG_STATUS_RX_DATA_TIMEOUT:
			//no data in READ_CMD_TIMEOUT, try again.
			//printf("Timeout on read.\n");
			break;

		case SANG_STATUS_BUFFER_TOO_SMALL:
			//Recieved data longer than the pre-configured maximum.
			//Maximum length is set in 'Interface Properties',
			//in the 'Device Manager'.
			cli_out(ERR, "Received data longer than buffer passed to API.\n");
			break;

		case SANG_STATUS_LINE_DISCONNECTED:
			//Front end monitoring is enabled and Line is
			//in disconnected state.
			//Check the T1/E1 line is in "Connected" state,
			//alse check the Alarms and the message log.
			cli_out(INFO, "%s: SANG_STATUS_LINE_DISCONNECTED\n", sang_if->device_name);
			Sleep(1000);//wait - to avoid busy loop
			break;

		case SANG_STATUS_RX_DATA_AVAILABLE:
			sang_if->callback_functions.got_rx_data(sang_if, &sang_if->rx_data);
			break;

		case SANG_STATUS_TDM_EVENT_AVAILABLE:
			sang_if->callback_functions.got_TdmApiEvent(sang_if, &sang_if->rx_data);
			break;

		default:
			cli_out(ERR, "%s: Rx: Unknown Operation Status: %d\n", 
				sang_if->device_name, pri->operation_status);
			break;
		}//switch()
	}//while()

	sang_if->terminate_tx_rx_threads = 2;

    cli_out(INFO, "\n%s: RxThreadFunc() - end\n", sang_if->device_name); 

    return 0; 
}

///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_ring_tone()
{
	cli_out(DBG_IFACE, "%s:start_ring_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_RING;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_ENABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}

int sangoma_interface::stop_ring_tone()
{
	cli_out(DBG_IFACE, "%s:stop_ring_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_RING;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_DISABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_congestion_tone()
{
	cli_out(DBG_IFACE, "%s:start_congestion_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_CONGESTION;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_ENABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}

int sangoma_interface::stop_congestion_tone()
{
	cli_out(DBG_IFACE, "%s:stop_congestion_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_CONGESTION;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_DISABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_busy_tone()
{
	cli_out(DBG_IFACE, "%s:start_busy_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_BUSY;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_ENABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}

int sangoma_interface::stop_busy_tone()
{
	cli_out(DBG_IFACE, "%s:stop_busy_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_BUSY;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_DISABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_dial_tone()
{
	cli_out(DBG_IFACE, "%s:start_dial_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_DIAL;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_ENABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}

int sangoma_interface::stop_dial_tone()
{
	cli_out(DBG_IFACE, "%s:stop_dial_tone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_TONE;

	tx_hdr->api_tx_hdr_event_tone = WP_API_EVENT_TONE_DIAL;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_DISABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::start_ringing_phone()
{
	cli_out(DBG_IFACE, "%s:start_ringing_phone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_RING;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_ENABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}

int sangoma_interface::stop_ringing_phone()
{
	cli_out(DBG_IFACE, "%s:stop_ringing_phone()\n", device_name);

	tx_hdr->api_tx_hdr_event_type = WP_API_EVENT_RING;

	tx_hdr->api_tx_hdr_event_mode = WP_API_EVENT_DISABLE;

	return wanpipe_api_ioctl(&wanpipe_api_cmd);
}
///////////////////////////////////////////////////////////////////////
int sangoma_interface::fxo_go_off_hook()
{
	cli_out(DBG_IFACE, "%s:fxo_go_off_hook()\n", device_name);

	if(wanif_conf_struct.sub_media != MOD_TYPE_FXO){
		cli_out(ERR, "%s: The 'Go Off Hook' command valid only for FXO!\n", device_name);
		return 1;
	}
	return tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_EVENT_TXSIG_OFFHOOK);
}

int sangoma_interface::fxo_go_on_hook()
{
	cli_out(DBG_IFACE, "%s:fxo_go_on_hook()\n", device_name);

	if(wanif_conf_struct.sub_media != MOD_TYPE_FXO){
		cli_out(ERR, "%s: The 'Go On Hook' command valid only for FXO!\n", device_name);
		return 1;
	}
	return tdmv_api_ioctl(&tdm_api_cmd, SIOC_WP_TDM_EVENT_TXSIG_ONHOOK);
}

int sangoma_interface::reset_interface_state()
{
	cli_out(DBG_IFACE, "%s:reset_interface_state()\n", device_name);

	if(wanif_conf_struct.sub_media == MOD_TYPE_FXS){
		stop_ring_tone();
		stop_congestion_tone();
		stop_busy_tone();
		stop_dial_tone();
		stop_ringing_phone();
	}

	if(wanif_conf_struct.sub_media == MOD_TYPE_FXO){
		fxo_go_on_hook();
	}

	return 0;
}
